/*
 * Jednoducha implementace ovladace pevneho disku pracujici
 * v rezimu PIO s adresaci LBA28.
 */

#include <stdint.h>

#include "io.h"
#include "util.h"
#include "ata.h"

#define ATA_SOFT_RESET_ATTEMPTS	(1000)

#define DEVICE_IO_BASE          (0x1f0) 
#define DEVICE_CONTROL_BASE     (0x3f6) 

#define DEVICE_CONTROL_REGISTER	(DEVICE_CONTROL_BASE + 0) /* control base + 0 */
#define DEVICE_IO_DATA          (DEVICE_IO_BASE + 0)
#define DEVICE_IO_STATUS        (DEVICE_IO_BASE + 7)

static int ata_soft_reset();

/**
 * Inicializuje zarizeni na primarni sbernici
 */
int ata_init()
{
	ata_soft_reset();
}

/**
 * Do ciloveho bufferu nacte sector_count sektoru.
 * Argument lba - udava adresu sektoru, odkud se bude cist.
 * Predpoklada se velikost sektoru 512 B.
 */
int ata_read(uint32_t lba, uint8_t sector_count, unsigned char *buf)
{
	uint8_t status;
	int slave = 0;
	outb(DEVICE_IO_BASE + 2, sector_count);
	outb(DEVICE_IO_BASE + 3, lba & 0xff);         /* LBAlo */
	outb(DEVICE_IO_BASE + 4, (lba >> 8) & 0xff);  /* LBAmid */
	outb(DEVICE_IO_BASE + 5, (lba >> 16) & 0xff); /* LBAhi */
	outb(DEVICE_IO_BASE + 6, ((lba >> 24) & 0x0f) | 0xe0 | slave << 4); /* LBAhi + 0xe0 nastaveni modu LBA + slave bit */
	outb(DEVICE_IO_BASE + 7, 0x20); /* read command */

	/* prvni pokusy o cteni stavu nemusi projit (zpozdeni 400ns) */
	for (int i = 0; i < 4; i++) {
		status = inb(DEVICE_IO_STATUS);
		if (status & 0x80) continue;   /* BSY */
		if (status & 0x08) break;      /* DRQ */
	}

	for (int sector = 0; sector < sector_count; sector++) {
		/* pred kazdym ctenim sektoru je potreba testovat dostupnost dat */
		do {
			status = inb(DEVICE_IO_STATUS);
		} while (status & 0x80); /* BSY */
		if (status & 0x21) return EATA_DEV_FAIL; /* ERR nebo DF */		

		/* prenos dat */
		for (int i = 0; i < 256; i++) {
			uint16_t data = inw(DEVICE_IO_DATA);
			*(buf++) = (data & 0xff);
			*(buf++) = (data >> 8) & 0xff;
		}
	}
	return EOK;
}

/**
 * Na adresu danou argumentem lba ulozi obsah buf.
 * Velikost zapisovanych dat je dana jako sector_count * 512 B.
 */

int ata_write(uint32_t lba, uint8_t sector_count, unsigned char *buf)
{
	uint8_t status;
	int slace = 0;
	outb(DEVICE_IO_BASE + 2, sector_count);
	outb(DEVICE_IO_BASE + 3, lba & 0xff);         /* LBAlo */
	outb(DEVICE_IO_BASE + 4, (lba >> 8) & 0xff);  /* LBAmid */
	outb(DEVICE_IO_BASE + 5, (lba >> 16) & 0xff); /* LBAhi */
	outb(DEVICE_IO_BASE + 6, ((lba >> 24) & 0x0f) | 0xe0 | slave << 4); /* LBAhi + 0xe0 nastaveni modu LBA + slave bit */
	outb(DEVICE_IO_BASE + 7, 0x30); /* write command */

	/* prvni pokusy o cteni stavu nemusi projit (zpozdeni 400ns) */
	for (int i = 0; i < 4; i++) {
		status = inb(DEVICE_IO_STATUS);
		if (status & 0x80) continue;   /* BSY */
	}

	for (int sector = 0; sector < sector_count; sector++) {
		/* pred kazdym zapisem sektoru je potreba testovat dostupnost */
		do {
			status = inb(DEVICE_IO_STATUS);
		} while (status & 0x80); /* BSY */
		if (status & 0x21) return EATA_DEV_FAIL; /* ERR nebo DF */		

		/* prenos dat */
		for (int i = 0; i < 256; i++) {
			uint16_t data = *buf | (*(buf + 1) << 8);
			outw(DEVICE_IO_DATA, data);
			buf += 2;
		}
	}
	return EOK;
}


/**
 * Vraci, jaky typ zarizeni je pripojeny na primarni sbernici
 */
int ata_detect(int slave)
{
	ata_soft_reset();
	outb(DEVICE_IO_BASE + 6, 0xa0 | slave << 4);

	/* zpozdeni 400 ns */
	inb(DEVICE_CONTROL_REGISTER);
	inb(DEVICE_CONTROL_REGISTER);
	inb(DEVICE_CONTROL_REGISTER);
	inb(DEVICE_CONTROL_REGISTER);

	/* ziska identifikaci (signaturu) zarizeni */
	unsigned sigl = inb(DEVICE_IO_BASE + 4);
	unsigned sigh = inb(DEVICE_IO_BASE + 5);

	if ((sigl == 0x14) && (sigh == 0xeb)) return ATA_PATAPI;
	if ((sigl == 0x69) && (sigh == 0x96)) return ATA_SATAPI;
	if ((sigl == 0x00) && (sigh == 0x00)) return ATA_PATA;
	if ((sigl == 0x3c) && (sigh == 0xc3)) return ATA_SATA;
	return ATA_UNKNOWN;
}


/**
 * Provede reset zarizeni na primarni sbernici, nutne pred zacatkem komunikace
 */
static int ata_soft_reset()
{
	outb(DEVICE_CONTROL_REGISTER, 0x06); /* nastav bit pro software reset, nIEN -- blokuj preruseni */
	outb(DEVICE_CONTROL_REGISTER, 0x02); /* vynuluj bit pro software reset, nIEN -- blokuj prerueseni */
	
	/* nekolikrat naprazdno zkusi precist stav
	 * specifikace vyzaduje zpozdeni 400ns */
	inb(DEVICE_CONTROL_REGISTER); 
	inb(DEVICE_CONTROL_REGISTER); 
	inb(DEVICE_CONTROL_REGISTER); 
	inb(DEVICE_CONTROL_REGISTER);

	for (int i = 0; i < ATA_SOFT_RESET_ATTEMPTS; i++) {
		uint8_t status = inb(DEVICE_CONTROL_REGISTER);	
		status &= 0xc0;       /* priznaky BSY a RDY */
		if (status == 0x40)   /* BSY = 0, RDY = 1   */
			return EOK;
	}
	return EATA_DEV_FAIL;
}
