/*
 *
 * Konfigurace systemu preruseni
 * 
 */

#include "irq.h"
#include "io.h"
#include "util.h"

/* tabulka IDT */
struct idt_entry idt_entries[256];

/**
 * Inicializuje obsluhu preruseni vcetne nastaveni jednotlivych vektoru
 */
void irq_init()
{
	for (int i = 0; i < 255; i++) {
		irq_set_handler(i, isr_0);
	}
	irq_set_handler(0x21, isr_keyb);

	irq_load_idt(255, idt_entries);

	pic_init();
	pic_enable_irq(PIC_MASTER, 0x01);
}

/**
 * Pro dany vektor preruseni nastavi odpovidajici handler
 */
void irq_set_handler(unsigned char irq, void *handler)
{
	idt_entries[irq].offset_1  = ((unsigned int) handler) & 0xffff;
	idt_entries[irq].offset_2  = (((unsigned int) handler) >> 16) & 0xffff;
	idt_entries[irq].selector  = 0x08; /* selektor kodoveho segmentu */
	idt_entries[irq].zero      = 0x00; /* vzdy 0 */ 
	idt_entries[irq].type_attr = 0x8e; /* 0x80 - present (je pouzit) | 0x0e -- 32bit mode + interrupt */
}

/**
 * Inicializuje PIC
 * https://pdos.csail.mit.edu/6.828/2005/readings/hardware/8259A.pdf
 */
void pic_init()
{
	/* ulozi nastavene masky */
	uint8_t mask_1 = inb(PIC1_DATA);
	uint8_t mask_2 = inb(PIC2_DATA);

	/* zahajeni inicializace */
	outb(PIC1_CMD, 0x11);
	outb(PIC2_CMD, 0x11);

	/* namapuje preruseni PIC (master) na vektory 32 az 39 */
	outb(PIC1_DATA, 0x20);

	/* namapuje preruseni PIC (slave) na vektory 40 az 47 */
	outb(PIC2_DATA, 0x28);

	/* nastavi pro master PIC, ze slave je preruseni IRQ2 (0000 0100) */
	outb(PIC1_DATA, 4);

	/* slave PIC je pripojeny jako IRQ2 (binarni cislo) */
	outb(PIC2_DATA, 2); 

	/* posledni konfiguracni slovo viz: https://wiki.osdev.org/PIC#Programming_the_PIC_chips */
	/* pouzije se rezim 8086/88 nikoliv 8080/85 */
    	outb(PIC1_DATA, 0x01); 
    	outb(PIC2_DATA, 0x01);

	/* obnovi nastavene masky */
	outb(PIC1_DATA, mask_1);
	outb(PIC2_DATA, mask_2);
}

/**
 * Na zvolenem PIC (PIC_MASTER/PIC_SLAVE) povoli preruseni dana maskou
 */
void pic_enable_irq(unsigned pic, uint8_t mask)
{
	uint8_t mask_old = inb(pic);
	outb(pic, mask_old | mask);
}

/**
 * univerzalni handler preruseni
 */
void irq_handler_generic(unsigned int z)
{
	kprint_color(12, 0);
	kprint("IRQ:");
	kprint_i(z);
}

/**
 * handler preruseni vyvolaneho klavesnici
 */
void irq_handler_keyb(unsigned int scan_code)
{
	kprint_color(3, 0);
	kprint("SCAN_CODE:");
	kprint_i(scan_code);
	kprint("\r\n");
}


